""" 
Docs::

    ### Install
        pip install --upgrade fire utilmy
        # pip install -r pip/pip_full.txt

        cd webapi/asearch
        mkdir -p ztmp


    ### Dataset Download
        cd asearch
        mkdir -p ./ztmp/
        cd "./ztmp/"
        git clone https://github.com/arita37/data2.git   data
        cd data
        git checkout text

       #### Check Dataset
       cd ../../
       ls ./ztmp/data/ner/ner_geo/




    ### Usage:
       cd asearch
       export PYTHONPATH="$(pwd)"
       alias pyclass="python nlp/cats/multilabel.py  "

       ### Data prep
            python nlp/cats/multilabel.py data_arxiv_create_norm  --dirin "./ztmp/data/cats/arxiv/raw" 


       ### Train
           pyclass run_train  --dirout ztmp/exp   --cfg config/traina/train1.yaml --cfg_name class_deberta_v1 


       pyclass run_infer  --dirout ztmp/exp   --cfg config/traina/train1.yaml --cfg_name class_deberta_v1 



    #### Dataset Test
        https://www.kaggle.com/datasets/spsayakpaul/arxiv-paper-abstracts

        For label = "cs.ML"
        Cat1 : cs
        Cat2:  ML



    ### Fine Tuning Strategies using LLM and SetFit

        https://medium.com/@xmikex83/using-large-language-models-to-train-smaller-ones-ee64ff3e4bd3

        https://huggingface.co/MoritzLaurer/deberta-v3-large-zeroshot-v2.0



"""
if "Import":
    import time,  json,re, os, pandas as pd, numpy as np,copy
    from dataclasses import dataclass
    from typing import Optional, Union
    from box import Box

    from collections import Counter
    from sklearn.metrics import f1_score    

    import datasets 
    from datasets import load_dataset, Dataset, load_metric
    import evaluate

    # If issue dataset: please restart session and run cell again
    from transformers import (
        BitsAndBytesConfig,
        GenerationConfig,
        HfArgumentParser,
        TrainingArguments,
        Trainer,

        ### 
        DataCollatorWithPadding,
        AutoTokenizer,

        ### Tokenizer / Sequence
        AutoModelForMultipleChoice,        
        AutoModelForSemanticSegmentation,
        AutoModelForSequenceClassification,

        ### LLM
        AutoModelForCausalLM,
        PhiForCausalLM,

    )
    # from transformers.models.qwen2.modeling_qwen2 import
    from transformers.tokenization_utils_base import PreTrainedTokenizerBase, PaddingStrategy
    from seqeval.metrics import f1_score, precision_score, recall_score, classification_report

    import spacy, torch
    # from pylab import cm, matplotlib

    from utilmy import (date_now, date_now, pd_to_file,log,pd_read_file, os_makedirs,
                        glob_glob, json_save, json_load, config_load,
                        dict_merge_into
    )

    from utils.util_exp import (exp_create_exp_folder, exp_config_override, exp_get_filelist)


#################################################################################################
######## Dataloder Customize  XXX ##############################################################
def data_arxiv_create_norm(dirin="./ztmp/data/cats/arxiv/raw"):
    """ 
         python nlp/cats/multilabel.py data_arxiv_create_norm  --dirin "./ztmp/data/cats/arxiv/raw" 

    """
    dirdata = dirin.split("raw")[0]

    flist = glob_glob(dirin)
    if len(flist)< 1: 
        from utilmy.util_download import donwload_kaggle
        url = "https://www.kaggle.com/datasets/spsayakpaul/arxiv-paper-abstracts"
        donwload_kaggle(url, dirout= dirdata + "/donwload" )


    df = pd_read_file(dirin, )
    log(df.head(5).T, df.columns, df.shape)
    cols = list(df.columns)
    log(cols, df.shape)
    assert df[[ "titles", "terms"  ]].shape


    ### train/test split
    cols = list(df.columns)
    if "cat1" not in cols :
        ##### Custom to Arxiv dataset
        df["terms"] = df["terms"].apply(lambda x:  x.replace("[", "" ).replace("]", "" ))   # ['cs.CV', 'cs.AI', 'cs.LG']

        df["terms"] = df["terms"].apply(lambda x:  x.replace(" ", "" ))   # ['cs.CV', 'cs.AI', 'cs.LG']

        df["label2"] = df["terms"].apply(lambda x:  sorted(x.split(",")) )  # ['cs.CV', 'cs.AI', 'cs.LG']
        df["cat1"] = df["label2"].apply(lambda x: x[0].split(".")[0])  # "cs"
        df["cat2"] = df["label2"].apply(lambda x: x[0].split(".")[-1])  # "CV"
        df["cat3"] = df["label2"].apply(lambda x: x[1].split(".")[0] if len(x)>1 else "NA" )  # "cs"
        df["cat4"] = df["label2"].apply(lambda x: x[1].split(".")[-1] if len(x)>1 else "NA")  # "AI"

        df["labels"] = df.apply(lambda x: x["cat1"] + "," + x["cat2"] + "," + x["cat3"] + "," + x["cat4"], axis=1)
        df["labels"] = df["labels"].apply(lambda x: x.replace("'", ""))

    df  = df.rename(columns={"titles": "text"}) 


    assert df[[   "text", "labels"  ]].shape
    n_train = int(len(df)*0.8)
    df_val  = df.iloc[n_train:,:]
    df      = df.iloc[:n_train,:]
    n_val   = len(df_val)

    pd_to_file(df,     f"{dirdata}/train/df_{n_train}.parquet", show=1)
    pd_to_file(df_val, f"{dirdata}/val/df_{n_val}.parquet",     show=1)


    log("############# Create Label Mapper")
    dlabel = LABELdata()
    dlabel.create_mapper(dirdata=dirdata + "/train/*.parquet", 
                         dirout=dirdata  + "/meta/meta.json" )
    log(dlabel.meta_dict)                     


def data_arxiv_load(dirin="./ztmp/data/cats/arxiv"):
    """ 
        In [140]: df.head(1).T                                                                0
        abstracts                                               None
        summaries  Stereo matching is one of the widely used tech...
        terms                                        'cs.CV','cs.LG'
        titles     Survey on Semantic Stereo Matching / Semantic ...
        label2                                    ['cs.CV', 'cs.LG']
        cat1                                                     'cs
        cat2                                                     CV'
        cat3                                                     'cs
        cat4                                                     LG'
        labels                                           cs,CV,cs,LG


    """ 
    df = pd_read_file(dirin + "/train", )
    log_pd(df)
    assert df[[ "text", "labels"  ]].shape


    df_val = pd_read_file(dirin + "/val", )
    log_pd(df_val)
    assert df_val[[ "text", "labels"  ]].shape

    assert len(df_val)>1 and df.shape[1]>= 2
    assert len(df)>1     and df.shape[1]>= 2 
    return df, df_val



def data_arxiv_load_label_mapper(dirmeta="./ztmp/data/cats/arxiv/meta/meta.json"):
    labeldata = LABELdata()
    # I2L, L2I, NLABEL_TOTAL, meta_dict =  dlabel.load_mapper(dirmeta=dirmeta)
    #return I2L, L2I, NLABEL_TOTAL, meta_dict 
    
    I2L, L2I, NLABEL_TOTAL, meta_dict =  labeldata.load_mapper(dirmeta=dirmeta)

    return labeldata, meta_dict 





########################################################################################
##################### Data Validator ###################################################
def log_pd(df,n=2):
    log(df.head(n),"/n", list(df.columns), "/n", df.shape)




#################################################################################################
######## Dataloder Common #######################################################################
def test():
    data = {'col1': ['a', 'd', 'g'], 'col2': ['b', 'e', 'h'], 'col3': ['c', 'f', 'i']}
    df = pd.DataFrame(data)

    dlabel = LABELdata()
    df = dlabel.pd_labels_merge_into_singlecol(df, cols=['col1', 'col2', 'col3'], colabels="colnew")
    log(df["colnew"].values[0])



class LABELdata:
    from utilmy import (date_now, date_now, pd_to_file,log,pd_read_file, os_makedirs,
                        glob_glob, json_save, json_load, config_load,
                        dict_merge_into)

    def __init__(self, dirdata=None, dirmeta=None):
        """ Label Data Storage and converter methods.

            dlabel = LABELdata()
            dlabel.create_mapper(dirin="./ztmp/data/cats/arxiv/train/df_8000.parquet")
            print(dlabel.I2L, dlabel.L2I, dlabel.NLABEL_TOTAL)

        """
        self.dirdata = dirdata   ###  training data raw files
        self.dirmeta = dirmeta   ###  meta.json file

        self.I2L, self.L2I,  self.meta_dict = {},{},{}
        self.I2CLASS, self.CLASS2I = {}, {}
        self.NLABEL_TOTAL = 0 


    def save_mapper(self, dirmeta=None):
        """ Save the json mapper to meta.json
        """
        dirout2 = dirmeta if dirmeta is not None else self.dirmeta 
        dirout2 = dirout2 if ".json" in dirout2 else dirout2 + "/meta.json"
        json_save(self.meta_dict, dirout2 )
        log(dirout2)


    def load_mapper(self, dirmeta:str=None):
        """Load the mapper from a directory containing meta.json 
        Args: dirmeta (str, optional): The directory containing the meta.json
        Returns: dict containing all the mapping.
        """

        dirmeta = dirmeta if dirmeta is not None else self.dirmeta
        flist = glob_glob(dirmeta)
        flist = [ fi for fi in flist if ".json" in fi.split("/")[-1]  ]
        fi = flist[0]

        if "json" in fi.split("/")[-1].split(".")[-1]:
            with open(fi, 'r') as f:
                meta_dict = json.load(f)

            if "meta_dict" in meta_dict.get("data", {}):
                ### Extract the meta_dict from config training
                meta_dict = meta_dict["data"]["meta_dict"] 

            self.NLABEL_TOTAL = meta_dict["NLABEL_TOTAL"]
            self.I2L = { int(ii): label   for ii, label in meta_dict["I2L"].items() } ## Force encoding
            self.L2I = { label  : int(ii) for label,ii  in meta_dict["L2I"].items() }

            self.dirmeta = fi
            return self.I2L, self.L2I, self.NLABEL_TOTAL, meta_dict
        else:
            log(" need meta.json")

    def create_mapper(self, dirdata="./ztmp/data/cats/arxiv/*.parquet",cols_class=None, dirout=None,
                      merge_all_labels_into_single_class=1):
        """Create a mapper json file for labels from raw training data
            python nlp/cats/multilabel.py labels_load_create_mapper --dirin "./ztmp/data/cats/arxiv/train/*.parquet" --dirout "./ztmp/data/cats/arxiv/meta/meta.json"

        Doc::
            Args: dirin (str, optional):  df[["labels"]] as string joined by ","

            Returns: meta.json
                tuple: A tuple containing the following mapper dictionaries:
                    - I2L (dict): dict mapping labels to their corresponding indices.
                    - L2I (dict): dict mapping indices to their corresponding labels.
                    - NLABEL_TOTAL (int): The total number of labels.
                    - meta_dict (dict): dict containing additional metadata about the labels.

        """          
        flist = glob_glob(dirdata)
        df = pd_read_file(flist)
        assert df[["labels"]].shape

        ##### Expand the column lbael  into multiple columns
        df2, cols_class = self.pd_labels_split_into_cols(df, colabel="labels", cols_class=cols_class)
        #df2 = df2.drop_duplicates()


        #####  mapping Label --> Index : Be careful of duplicates
        I2L,L2I          = {}, {}
        I2CLASS, CLASS2I = {}, {}
        dd  = {"labels_n_unique":{},  "labels" : {   }  }
        idx = -1
        for class_i in cols_class:
            vv  = df2[class_i].drop_duplicates().values
            dd["labels_n_unique"][class_i] = len(vv)
            dd["labels"][class_i] = { "labels": list(vv) , 
                                      "freq":  df2[class_i].value_counts().to_dict()  }

            class_i2 = f"{class_i}"  
            if int(merge_all_labels_into_single_class) == 1:
                log("merge_all_labels_into_single_class") 
                class_i2 = ""
                
            ### For each label of the column, add the indexesm add to the class indexes. 
            CLASS2I[class_i2] =[]
            for  label in vv:
                label  = label.strip()
                label2 = f"{class_i2}-{label}" if class_i2 != "" else label   ### gender-male, gender-female
                # log("label:", label2)
                if label2 not in L2I:
                    idx += 1
                    L2I[label2] = idx                ### red --> 1, blue --> 2
                    I2L[idx]    = label2             ### 1 --> red, 2 --> blue

                    I2CLASS[idx]= class_i2           ####  red --> color, blue --> color
                    CLASS2I[class_i2].append( idx)   #### Color --> [red, blue]
                    
        #### All classes
        dd["class_cols"]   = list(cols_class)  ### List of class columns
        dd["NCLASS_TOTAL"] = len(cols_class)

        #### Labels of each class
        NLABEL_TOTAL = len(L2I) + 1  ### + 1 for NA/Unknown Values        
        dd["NLABEL_TOTAL"] = NLABEL_TOTAL
        dd["I2L"] = I2L
        dd["L2I"] = L2I
        dd["I2CLASS"] = I2CLASS   ####  red --> color, blue --> color
        dd["CLASS2I"] = CLASS2I   ####  Color --> [red, blue]

        self.meta_dict = dd
        self.I2L = I2L 
        self.L2I = L2I 
        self.NLABEL_TOTAL = NLABEL_TOTAL 


        if dirout is not None:
           self.save_mapper(dirout) 

        return I2L, L2I, NLABEL_TOTAL, self.meta_dict 

    
    def pd_labels_split_into_cols(self, df, colabel="labels", cols_class=None):
        df = df[[colabel]]
        if isinstance(df[colabel].values[0], str):
            df[colabel] = df[colabel].apply(lambda xstr: [t.strip() for t in xstr.split(",")])
        else:
            df[colabel] = df[colabel].apply(lambda xlist: [t.strip() for t in xlist])

        ncols = len(df[colabel].values[0])
        cols_class = [f"class_{i}" for i in range(ncols)] if cols_class is None else cols_class
        df2 = pd.DataFrame(df[colabel].tolist(), columns=cols_class)
        df2.index = df.index

        return df2, cols_class

    def pd_labels_merge_into_singlecol(self, df, cols=None, sep=",", colabels="labels"):
        if cols is None:
            cols = df.columns.tolist()
        df[colabels] = df[cols].astype(str).agg(sep.join, axis=1)
        return df

    def to_onehot(self, xstr:str):
        """ Converts a string of ","labels into a one-hot encoded list.
        """
        zero, one = 0.0, 1.0  ## HFace needs float !!

        labels_onehot = [float(zero) ] * self.NLABEL_TOTAL
        for label in xstr.split(","):    
            ### If mising, One Label INDEX for unknown
            label_id = self.L2I.get(label, self.NLABEL_TOTAL-1 )  
            # log("label_id:",label,  label_id)
            labels_onehot[label_id] = one #

        return labels_onehot



def data_tokenize_split(df, tokenizer, labelEngine, cc):
    """ 

        {'input_ids': [[1, 8382, 277, 39093, 25603, 31487, 840, 39093, 28368, 59543, 2, 2, 2, 2, 2, 2, 2, 2, 2], 
                      [1, 14046, 271, 5203, 473, 13173, 75204, 270, 6547, 40457, 267, 13946, 5648, 2, 2, 2, 2, 2, 2]],

        'token_type_ids': [[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], 
                           [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], 

        'attention_mask': [[1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0], 
                            [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], 
                            [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0]], 

        'labels': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 

    """
    cols         = list(df.columns)
    max_length   = cc.data.sequence_max_length
    NLABEL_TOTAL = cc.data.nlabel_total
    log("nlabel_total: ", NLABEL_TOTAL)


    df["labels"] = df["labels"].apply(lambda x: labelEngine.to_onehot(x) )
    # df["labels"] = df["labels"].apply(lambda x: to_onehot(x) )
    log( df[[ "text", "labels" ]].head(1).T )
 

    def preprocess_func(row):
        #### Text tokenizer
        out = tokenizer(row["text"], truncation=True, padding=True, max_length= max_length,
                   return_offsets_mapping= False,
                   return_overflowing_tokens=False)

        out["labels"] = row["labels"]     
        # log(out)    
        # output["input_ids"] = output.pop("input_ids")  # Add input_ids to  output
        return out

    #### Encode texts
    ds = Dataset.from_pandas(df[["text", "labels"  ]])
    ds = ds.map(preprocess_func, batched=True)


    #### Filter labels with only a single instance
    label_counts = Counter([tuple(label) for label in ds["labels"]])
    valid_labels = [label for label, count in label_counts.items() if count > 1]
    ds           = ds.filter(lambda row: tuple(row["labels"]) in valid_labels)

    #### Reduce columns in dataset
    ds = ds.remove_columns(['text'  ] )

    #  ['text', 'labels', 'input_ids', 'token_type_ids', 'attention_mask']
    # ds           = ds.remove_columns( cols)
    # ds = ds.remove_columns(['overflow_to_sample_mapping', 'offset_mapping', ] + cols)
    log(ds)
    return ds





########################################################################################
##################### Tokenizer helper ################################################
def DataCollatorClassification(tokenizer=None):
    from transformers import DataCollatorWithPadding
    data_collator = DataCollatorWithPadding(tokenizer=tokenizer, padding="longest")
    return data_collator 




#######################################################################################
##### Define A model with custom classifier head : 3 softmax
class MultiSoftmaxClassifier(AutoModelForSequenceClassification):
    def __init__(self, config):
        """
        3 softmax head : one per class group: 
           class A  : A1-A5, 
           classs B : B1-B4, 
           class  C : C1-C2

        ### Usage code:
            # Load the model and tokenizer
            modelid   = 'microsoft/deberta-v3-base'
            tokenizer = AutoTokenizer.from_pretrained( modelid)
            model     = MultiSoftmaxClassifier.from_pretrained(modelid, num_labels=11)
            model.eval()  # Set the model to evaluation mode

            # Tokenize the input text
            inputs = tokenizer(texts, padding=True, truncation=True, return_tensors="pt")
            
            # Move inputs to the same device as model
            inputs = {k: v.to(model.device) for k, v in inputs.items()}
            
            # Perform inference
            with torch.no_grad():
                outputs = model(**inputs)
            print( probabilities.cpu().numpy() )

        """
        super().__init__(config)
        self.classifier = torch.nn.Linear(config.hidden_size, config.num_labels)  # Total labels across all classes

    def forward(self, **inputs):
        outputs = super().forward(**inputs)
        # logits = outputs.logits
        logits = self.classifier(outputs.pooler_output)  # Use the classifier on the pooled output

        # Apply softmax per class group: A1-A5, B1-B4, C1-C2
        # 1 Class can only contain 1 label only,
        logits = torch.cat([
            torch.nn.functional.softmax(logits[:, :5],  dim=1),
            torch.nn.functional.softmax(logits[:, 5:9], dim=1),
            torch.nn.functional.softmax(logits[:, 9:],  dim=1)
        ], dim=1)
        return torch.nn.functional.softmax(logits, dim=1)


def test5():
    import torch
    from transformers import AutoTokenizer

    # Load the model and tokenizer
    modelid   = 'microsoft/deberta-v3-base'
    tokenizer = AutoTokenizer.from_pretrained( modelid)
    model     = MultiSoftmaxClassifier.from_pretrained(modelid, num_labels=11)
    model.eval()  # Set the model to evaluation mode

    # Function to perform inference
    def predict(texts):
        # Tokenize the input text
        inputs = tokenizer(texts, padding=True, truncation=True, return_tensors="pt")
        
        # Move inputs to the same device as model
        inputs = {k: v.to(model.device) for k, v in inputs.items()}
        
        # Perform inference
        with torch.no_grad():
            outputs = model(**inputs)
        
        # Apply softmax to logits to get probabilities
        logits = outputs['logits']
        probabilities = torch.nn.functional.softmax(logits, dim=1)
        
        # Convert probabilities to numpy array
        probabilities = probabilities.cpu().numpy()
        
        return probabilities

    # Example usage
    texts = ["Sample text for classification."]
    predictions = predict(texts)
    print("Predicted probabilities:", predictions)



#################################################################################
######################## Train ##################################################
def run_train(cfg=None, cfg_name="class_deberta", dirout="./ztmp/exp", istest=1):
    """ 
       python nlp/cats/multilabel.py run_train  --dirout ztmp/exp   --cfg config/traina/train1.yaml --cfg_name class_deberta 

      Multi class Multi Labels   ( type of text, author category, ...)
       Class columnns
              
            classA,       classB,        classC,      classD

             LA_1,        LB_2
             LA_2,        LB_3  
             LA_3,        LB_3          ...      ...
             
       Simpligy notation:  merge by ","

              labels ","

                 LA_1,LB_3,LC_9
                 LA_9,LB_3,LC_4
                 ....

                 LA_4,LB_1,LC_12

        Labels --> list of string.


    """  
    log("\n###### Config Load   #############################################")
    cfg0 = config_load(cfg)
    cfg0 = cfg0.get(cfg_name, None) if cfg0 is not None else None


    log("\n###### User default Params   #######################################")
    if "config":
        cc = Box()
        cc.model_name='microsoft/deberta-v3-base'

        #### Data name
        cc.dataloader_name = "data_arxiv_load"
        cc.datamapper_name = "data_arxiv_load_label_mapper"

        cc.n_train = 6  if istest == 1 else 1000000000
        cc.n_val   = 2 if istest == 1 else 1000000000

        ############# Train Args
        aa = Box({})
        aa.output_dir                  = f"{dirout}/log_train"
        aa.per_device_train_batch_size = 64
        aa.gradient_accumulation_steps = 1
        aa.optim                       = "adamw_hf"
        aa.save_steps                  = min(100, cc.n_train-1)
        aa.logging_steps               = min(50,  cc.n_train-1)
        aa.learning_rate               = 1e-5
        aa.max_grad_norm               = 2
        aa.max_steps                   = -1
        aa.num_train_epochs            = 1
        aa.warmup_ratio                = 0.2 # 20%  total step warm-up
        # lr_schedulere_type='constant'
        aa.evaluation_strategy = "epoch"
        aa.logging_strategy    = "epoch"
        aa.save_strategy       = "epoch"
        cc.hf_args_train = copy.deepcopy(aa)

        ############# Model Args
        cc.hf_args_model = {}
        cc.hf_args_model.model_name   = cc.model_name
        cc.hf_args_model.problem_type = "multi_label_classification"


        #### Override by Yaml config  #####################################
        cc = exp_config_override(cc, cfg0, cfg, cfg_name)


    log("\n###### Experiment Folder   #########################################")
    cc = exp_create_exp_folder(task="class_deberta", dirout=dirout, cc=cc)
    log(cc.dirout);  del dirout


    log("\n###### Model : Training params ######################################")
    args = TrainingArguments( ** dict(cc.hf_args_train))


    log("\n###### User Data Load   #############################################")
    from utilmy import load_function_uri
    ### String names --> Create Python function object
    dataloader_fun = load_function_uri(cc.dataloader_name, globals())
    datamapper_fun = load_function_uri(cc.datamapper_name, globals())


    df, df_val             = dataloader_fun()  ## = dataXX_load_prepro()
    labelEngine, meta_dict = datamapper_fun()  ## Label to Index, Index to Label
    #L2I, I2L, NLABEL_TOTAL, meta_dict = datamapper_fun()  ## Label to Index, Index to Label

    df, df_val = df.iloc[:cc.n_train,:], df_val.iloc[:cc.n_val,:]
    log( df_val.shape)

    if "params_data":
        cc.data ={}
        cc.data.cols          = df.columns.tolist()
        cc.data.cols_required = ["text", "labels" ]
        # cc.data.cols_remove   = ['overflow_to_sample_mapping', 'offset_mapping', ] + columns
        # cc.data.nclass        = NCLASS  ### Number of NER Classes.
        cc.data.meta_dict     = meta_dict

        NLABEL_TOTAL = cc.data.meta_dict['NLABEL_TOTAL']
        I2L = cc.data.meta_dict['I2L']   ### label to Index Dict
        L2I = cc.data.meta_dict['L2I']   ### Index to Label Dict     
        cc.data.nlabel_total        = NLABEL_TOTAL  ### Number of NER Classes.
        cc.data.sequence_max_length = 128

        cc.hf_args_model.num_labels = NLABEL_TOTAL ## due to BOI notation

 
    log("\n###### Dataloader setup  #######################################")
    tokenizer = AutoTokenizer.from_pretrained(cc.model_name)
    tokenizer.pad_token = tokenizer.eos_token

    dataset_train = data_tokenize_split(df,     tokenizer, labelEngine, cc, )    
    dataset_valid = data_tokenize_split(df_val, tokenizer, labelEngine, cc,)    


    data_collator = DataCollatorClassification(tokenizer=tokenizer)
    batch = data_collator([dataset_train[0], dataset_train[1]])
    log(batch)


    log("\n######### Model : Init #########################################")
    model = AutoModelForSequenceClassification.from_pretrained(cc.model_name, 
                               num_labels= cc.hf_args_model.num_labels,  ## Check if this correct ????
                               id2label= I2L, 
                               label2id= L2I,
                               problem_type = "multi_label_classification")

    # for i in model.deberta.parameters():
    #   i.requires_grad=False

    log("\n######### Model : Training start ##############################")
    trainer = Trainer(model, args,
        train_dataset= dataset_train,
        eval_dataset = dataset_valid,
        data_collator= data_collator,
        tokenizer    = tokenizer,
        compute_metrics=metrics_callback_train,
    )

    json_save(cc, f'{cc.dirout}/config.json')
    trainer_output = trainer.train()
    trainer.save_model( f"{cc.dirout}/model")

    cc['metrics_trainer'] = trainer_output.metrics
    log(str(cc)[:100])
    json_save(cc, f'{cc.dirout}/meta.json', show=0)
    json_save(cc, f'{cc.dirout}/model/meta.json', show=0)  ### Required when reloading for inference


    log("\n######### Model : Eval Predict  ######################################")
    preds_proba, labels, _ = trainer.predict(dataset_valid)
    df_val                 = pd_predict_format(df_val, preds_proba, labels, cc)
    assert df_val[[ "text", "labels", "pred_labels", "pred_proba"     ]].shape
    pd_to_file(df_val, f'{cc.dirout}/dfval_pred_labels.parquet', show=1)  


    log("\n######### Model : Eval Metrics #######################################")    
    assert df_val[ [ "text", "labels", "pred_labels", "pred_proba"  ]].shape
    metrics_calc_full(df_val, metric_list=["accuracy", "f1", "precision", "recall"], 
                      dirout = f'{cc.dirout}/dfval_pred_metrics.csv' )


def pd_predict_format(df_val:pd.DataFrame, preds_proba:list, labels:list, cc )->pd.DataFrame:
    """ preds_proba: (batch, seq, NCLASS*2+1)    [-3.52225840e-01  5.51

        labels:      (batch,seq)   [[ 0   0   1   1   0    1   1    0   0   0   ]
                                    [ 0   0   1   1   0    1   1    0   0   0   ]]
        pred_labels: (batch, seq)  [[ 1  0  0  1   0  1  0  1  1  0  1 ]
                                    [ 1  0  0  1   0  1  0  1  1  0  1 ] ]

     
     BE CAREFUL Many  cases :

         Softmax :  1 single Label per prediction class:  np.argmax(preds_proba, axis=-1)
  
         Sigmoid :  Manay Label per prediction class :     1 if proba > 0.5 else 0

    """
    def labels_calc_softmax():
       ### SoftMax : 1 label for all of the labels ( only one  1  per list)
       pred_labels = np.argmax(preds_proba, axis=-1)
       return pred_labels


    def labels_calc_softmax_multi(label_idx_list):
       ### SoftMax : 1 label for for each class group, 
       pred_labels = []
       for label_idx in label_idx_list:
          pi = np.argmax(preds_proba[label_idx ], axis=-1)
          pred_labels.append(pi)  

       return pred_labels


    def labels_calc_sigmoid(proba_threshold=0.5):
       ### Multi-label, mutli class : Each Label is Indepedent
       pred_labels = [  [ 1 if pi > proba_threshold else 0 for pi in yproba_list    ] for yproba_list in list(preds_proba) ]
       return pred_labels


    pred_labels = labels_calc_sigmoid(proba_threshold=0.5)

    log("\npred_proba: ", str(preds_proba)[:50],)
    log("labels: ",       str(labels)[:50]      )
    log("pred_class: ",   str(pred_labels)[:50], "\n\n" )

    df_val['pred_proba']  = list(preds_proba) 
    df_val['pred_labels'] = list(pred_labels) ### 2D array into List of List  ## ValueError: Expected a 1D array, got an array with shape (2, 25)

    #df_val['pred_class_list_records'] = pd_predict_convert_class_to_records(df_val,)
    return df_val



################################################################################
########## Metrics Helper    ###################################################
def metrics_callback_train(p):
    from sklearn.metrics import f1_score, accuracy_score, precision_recall_fscore_support

    def sigmoid(x):
       return 1 / (1 + np.exp(-x))

    pthreshold = 0.5 
    ##### Softmax
    #preds = p.predictions.argmax(-1)

    ##### sigmoid   
    preds_2d =[]
    for logit_list in p.predictions:
       preds_2d.append( [  1 if sigmoid(logit) > pthreshold else 0 for logit in logit_list  ] )

    labels_2d = p.label_ids
    
    dd = {}
    dd["accuracy"] = accuracy_score(labels_2d, preds_2d)
    dd["precision"], dd["recall"], dd["f1"], support = precision_recall_fscore_support(labels_2d, preds_2d, average='weighted')

    return dd



def metrics_calc_full(df, metric_list=None, dirout=None, cc=None):
    """ 
       Calcualte metrics per column class using numpy, sklearn.
          ztmp/exp/20240524/214636-class_deberta-6/dfval_pred_labels.parquet

       dirin = "ztmp/exp/20240524/214636-class_deberta-6/dfval_pred_labels.parquet"
       df    = pd_read_file(dirin)
       dfm   = metrics_calc_full(df, metric_list=["accuracy", "f1"], dirout=None)           

    """
    from utils.util_metrics  import metrics_eval 

    def np_to_single_array(array_array ):
        return np.array([list(arr) for arr in array_array ])

    metrics_list = ["accuracy", "f1", "precision", "recall"] if metric_list is None else metric_list

    pred_2d = np_to_single_array( df["pred_labels"].values)
    true_2d = np_to_single_array( df["labels"].values ) 
    pred_proba_2d = np_to_single_array( df["pred_proba"].values)    
    assert pred_2d.shape == true_2d.shape


    n_class = len(pred_2d[0])
    dfall   = pd.DataFrame()  ### Metrics
    for icol in range(n_class):
        dfi = metrics_eval(ypred = pred_2d[:,icol], 
                           ytrue = true_2d[:,icol], 
                           ypred_proba = pred_proba_2d[:,icol],
                           metric_list = metrics_list)

        #### Need to map One Hot colid to class name:
        #  

        dfi["class_name"] = f"onehot_{icol}"
        dfall = pd.concat((dfall, dfi ) ) 

    dfall.index = np.arange(0, len(dfall))
    dfall = dfall[[ "class_name",  "metric_name", "metric_val", "n_sample" ]]
     
    if dirout is not None:
        pd_to_file(df, dirout, show=1)    

    return dfall





################################################################################
########## Run Inference  ######################################################
def run_infer_batch(cfg:str=None,  cfg_name    = "cats_multilabel_infer",
                dirmodel    = "ztmp/models/gliclass/small", 
                dirdata     = "ztmp/data/text.csv",
                coltext     = "text",
                dirout      = "ztmp/data/class/predict/",
                device="cpu",
                batch_size  = 8,
                nmax= 10,
                  ):
   """Run prediction using a pre-trained  Deberta model.

    ### Usage
      alias pyclass="python nlp/cats/multilabel.py "
      dirmodel="./ztmp/exp/20240525/174321-class_deberta-6/model"

         pyclass run_infer_batch --dirmodel $dirmodel  --dirdata "ztmp/data/cats/arxiv/val/*"  --dirout ztmp/out/class/deberta_arxiv/

         pyclass run_infer_batch --cfg config/train.yaml     --cfg_name "cats_deberta_infer_v1"

    Args:
        cfg (dict)    : Configuration dictionary (default is None).
        dirmodel (str): path of pre-trained model 
        dirdata (str) : path of input data 
        coltext (str) : Column name of text
        dirout (str)  : path of output data 


      #log(model.predict("My name is John Doe and I love my car. I bought a new car in 2020."))
      #df["class_list_pred"] = df[coltext].apply(lambda x: model.predict(x, flat_class=True, threshold=0.5, multi_label=False))

   """
   from transformers import pipeline

   log("\n#### User config   ####################################################")
   cfg0 = config_load(cfg, )
   cfg0 = cfg0.get(cfg_name, {}) if isinstance(cfg0, dict) else {}

   dirmodel = cfg0.get("dirmodel", dirmodel)
   dirdata  = cfg0.get("dirdata",  dirdata)
   dirout   = cfg0.get("dirout",   dirout)
   dirmeta    = dirmodel  + "/meta.json"
   batch_size = cfg0.get("batch_size", batch_size)

   task="text-classification"
   # device = torch_get_device(device)

   log("\n#### Model pre-trained config   #######################################")
   cc = config_load(dirmeta)
   log(str(cc)[:80])
   class_cols:list = cc.data.get("class_cols", None)
   class_cols      = ["gender", "color", "style",  ] if class_cols is None else class_cols


   log("\n#### Model Load  ######################################################")
   ###  tasks are [ 'conversational',  'document-question-answering', 'feature-extraction', 'fill-mask', 'mask-generation', 'ner', 'object-detection', 'question-answering', 'sentiment-analysis', 'summarization', 'table-question-answering', 'text-classification', 'text-generation', 'text-to-audio', 'text-to-speech', 'text2text-generation', 'token-classification', 'translation',  'vqa', 'zero-shot-audio-classification', 'zero-shot-classification', 'translation_XX_to_YY']"
   pipe = pipeline(task=task, model=dirmodel, tokenizer=dirmodel, device=device)
   log(str(pipe)[:80])


   log("\n#### Inference Start loop  ############################################")
   flist = exp_get_filelist(dirdata)
   for ii,fi in enumerate(flist) :
      df = pd_read_file(fi, nrows= nmax)
      log(ii, fi,  df.shape)

      ### {'label': 'DC', 'score': 0.5918803811073303} per row
      pred_list_records = pipe( df[coltext].tolist(), batch_size= batch_size )
      df["pred_list_records"] = pred_list_records

      #assert df[[ "text", "labels", "pred_labels", "pred_proba"     ]].shape
      pd_to_file(df, f'{dirout}/df_pred_labels_{ii}.parquet', show=1)  
   log("\n#### Inference End ##################################################")



def run_eval2(cfg:str=None, 
                cfg_name    = "cats_gliclass_predict",
                dirmodel    = "ztmp/models/gliclass/small", 
                dirdata     = "ztmp/data/text.csv",
                coltext     = "text",
                dirout      = "ztmp/data/class/predict/",
                device="cpu"
                  ):
   """Run prediction using a pre-trained  Deberta model.

    ### Usage
      alias pyclass="python nlp/cats/multilabel.py "


      dirmodel="./ztmp/exp/20240525/174321-class_deberta-6/model"

      pyclass run_infer_batch --dirmodel $dirmodel  --dirdata "ztmp/data/cats/arxiv/val/*"  --dirout ztmp/out/class/deberta_arxiv/

      pyclass run_infer_batch --cfg config/train.yaml     --cfg_name "cats_deberta_infer_v1"

    Args:
        cfg (dict)    : Configuration dictionary (default is None).
        dirmodel (str): path of pre-trained model 
        dirdata (str) : path of input data 
        coltext (str) : Column name of text
        dirout (str)  : path of output data 


      #log(model.predict("My name is John Doe and I love my car. I bought a new car in 2020."))
      #df["class_list_pred"] = df[coltext].apply(lambda x: model.predict(x, flat_class=True, threshold=0.5, multi_label=False))

   """
   cfg0 = config_load(cfg, )
   cfg0 = cfg0.get(cfg_name, {}) if isinstance(cfg0, dict) else {}

   dirmodel = cfg0.get("dirmodel", dirmodel)
   dirdata  = cfg0.get("dirdata",  dirdata)
   dirout   = cfg0.get("dirout",   dirout)
   dirmeta = dirmodel  + "/meta.json"


   log("\n### Inference config   #####################################################")
   cc = config_load(dirmeta)
   log(str(cc)[:80])

   class_cols:list = cc.data.get("class_cols", None)
   class_cols      = ["gender", "color", "style",  ] if class_cols is None else class_cols


   log("\n### Model Load  ############################################################")
   model     = AutoModelForSequenceClassification.from_pretrained(dirmodel,)
   tokenizer = AutoTokenizer.from_pretrained(cc.model_name)
   tokenizer.pad_token = tokenizer.eos_token

   labelEngine = LABELdata()
   labelEngine.load_mapper(dirmeta= dirmeta)
   log(str(model)[:20])

   def preprocess_func(row):
        #### Text tokenizer
        out = tokenizer(row["text"])

        return out

   model.eval()
   flist = exp_get_filelist(dirdata)
   for ii,fi in enumerate(flist) :
     df = pd_read_file(fi)
     log(ii, fi,  df.shape)


     pred_list = torch_model_predict(model, df["text"].values, device="cpu", tokenizer=tokenizer, show=1)
     df["pred_labels"] = pred_list

     # dataset_valid = data_tokenize_split(df, tokenizer=tokenizer,labelEngine=labelEngine, cc=cc ) 
     # ds = Dataset.from_pandas(df[["text"  ]])
     # ds = ds.map(preprocess_func, batched=True, return_tensors="pt")
     # encoded_input = tokenizer(text)
     # encoded_input["input_ids"] = encoded_input.pop("input_ids")  # Add input_ids to  encoded_input
     # encoded_input = {k: v.to(device) for k, v in encoded_input.items()}
     # encoded_input = {k: v.to(device) for k, v in ds.items()}
     #X= [ tokenizer(txt, return_tensors="pt") for txt in df["txt"].values       ]

     #preds_proba, labels, _ = model(**X)
     df                     = pd_predict_format(df, preds_proba, labels, cc=cc)
     assert df[[ "text", "labels", "pred_labels", "pred_proba"     ]].shape
     pd_to_file(df, f'{dirout}/df_pred_labels_{ii}.parquet', show=1)  



def run_eval(dirmodel, device="cpu", dirdata= ".ztmp/data/arxiv_data/eval/" ,dataloader_name="dataset_arxiv", topk=3):
    """Eval

    python multilabel.py --dirmodel "./ztmp/pretrained/" --dirdata ".ztmp/data/arxiv_data/eval/"

    Args:
        dirmodel (str):  directory path of  model to be evaluated.
        device (str, optional):  device on which to run  evaluation. Defaults to "cpu".
        dirdata (str, optional):  directory path of  data for evaluation. Defaults to ".ztmp/data/arxiv_data/eval/".
        dataloader_name (str, optional):  name of  dataloader. Defaults to "dataset_arxiv".
        topk (int, optional):  number of top labels to consider. Defaults to 3.
    """
    model = DebertaForSequenceClassification.from_pretrained(dirmodel)

    log("##### Load Data")
    data, class_info, n_class_total        = data_load_convert_todict(dirin= dirdata, fun_name=dataloader_name,)
    train_dataset, test_dataset, tokenizer = data_tokenize_split(data)
    class_names = class_info["class_name"]


    log("##### Inference start")
    model.eval()
    text_list = [example["text"] for example in test_dataset]
    pred_cat  = []
    preds_lsit = torch_model_predict(model, text_list, device="cpu", tokenizer=None, show=1) 


def torch_model_predict(model, text_list, device="cpu", tokenizer=None, show=1):
    pred_list =[]
    for text in text_list:
        encoded_input = tokenizer(
            text, return_tensors="pt", padding=True, truncation=True, max_length=128
        )
        encoded_input["input_ids"] = encoded_input.pop("input_ids")  # Add input_ids to  encoded_input

        encoded_input = {k: v.to(device) for k, v in encoded_input.items()}

        output = model(**encoded_input)
        preds  = torch.sigmoid(output.logits).squeeze().detach().numpy()

        pred_list.append(preds)

        #top_labels  = np.argsort(preds)[-topk:]  # Get  top N labels
        #pred_labels = [class_names[i] for i in top_labels]  # Get  corresponding labels
        # pred_cat.append(pred_labels)

    if show >0 :
      #### Print abstracts with predicted labels
      for text, preds in zip(text_list, pred_list):
        print(f"Input: {text}\nPrediction: {preds}\n")
        break

    return pred_list




########################################################################################
####### Tooling ########################################################################
def data_load_convert_todict(dirin=".ztmp/arxiv_data.csv", fun_name="dataset_arxiv", label_nunique=None):
    """  Create custom category from raw label
         ['cs.CV', 'cs.AI', 'cs.LG']

    """ 
    from utilmy import load_function_uri
    dataset_loader_fun = globals()[fun_name]
    df, meta_dict = dataset_loader_fun(dirin)


    label_nunique = meta_dict.get("label_nunique")  ## number of unique categories per label
    if label_nunique is None:
        label_nunique = {"cat1": 3, "cat2": 5, "cat3": 6, "cat4": 7, }


    ####### Big OneHot Encoded  ######################################
    colslabel     = [  key for key in label_nunique.keys()]       
    n_class_total = sum([  label_nunique[key] for key in label_nunique.keys()])

    for collabel in colslabel:    
       df[ f"{collabel}_onehot"] = cat_to_onehot(df[collabel].values , ndim= label_nunique[collabel])

    df["cat_all_onehot"] = df.apply(lambda x :  sum( [x[ci+"_onehot"] for ci in colslabel ] )  ,axis=1 )


    # Convert DataFrame to a dictionary
    data = {
        "text":    df["text"].tolist(),
        "labels":  df["cat_all_onehot"].values.tolist(),
    }
    return data, label_nunique, n_class_total




def cat_to_onehot(cat_val:list, ndim=None):
    """ 
        # Example usage
        categories = ['2', '0', '1', '2']  # Example category indices as strings
        ndim = 3  # Number of unique categories
        encoded = encode_categories(categories, ndim)
        print(encoded)

    """
    from sklearn.preprocessing import OneHotEncoder
    import numpy as np    

    ndim= ndim if ndim is not None else len(np.unique(cat_val))

    # Ensure  input is an array
    categories = np.array(cat_val).reshape(-1, 1)
    
    # Create  OneHotEncoder with  specified number of dimensions
    encoder = OneHotEncoder(categories=[np.arange(ndim)], sparse=False)    
    onehot = encoder.fit_transform(categories)
    return onehot



def pd_predict_convert_class_to_records(df_val:pd.DataFrame, offset_mapping: list,
                                       col_classlist="pred_class_list", col_text="text")->pd.DataFrame:
    """Convert predicted classes into span records for NER.

    Args:
        df_val (pd.DataFrame): DataFrame containing the input data. It should have the following columns:
            - col_classlist (str): Column name for the predicted classes.
            - col_text (str): Column name for the text.
        offset_mapping (list): List of offset mappings.

    Returns:
        list: List of span records for NER. Each span record is a dictionary with the following keys:
            - text (str): text.
            - class_list (list): List of named entity records. Each named entity record is a dictionary with the following keys:
                - type (str)            : type of the named entity.
                - predictionstring (str): predicted string for the named entity.
                - start (int)           : start position of the named entity.
                - end (int)             : end position of the named entity.
                - text (str)            : text of the named entity.

    """
    # print(df_val)
    # assert df_val[[ "text", "input_ids", "offset_mapping", "pred_class_list"  ]]
    def get_class(c):
        if c == NCLASS*2: return 'Other'
        else: return I2L[c][2:]

    def pred2span(pred_list, df_row, viz=False, test=False):
        #     example_id = example['id']
        n_tokens = len(df_row['offset_mapping'][0])
        #     print(n_tokens, len(example['offset_mapping']))
        classes = []
        all_span = []
        for i, c in enumerate(pred_list.tolist()):
            if i == n_tokens-1:
                break
            if i == 0:
                cur_span = df_row['offset_mapping'][0][i]
                classes.append(get_class(c))
            elif i > 0 and (c == pred_list[i-1] or (c-NCLASS) == pred_list[i-1]):
                cur_span[1] = df_row['offset_mapping'][0][i][1]
            else:
                all_span.append(cur_span)
                cur_span = df_row['offset_mapping'][0][i]
                classes.append(get_class(c))
        all_span.append(cur_span)

        text = df_row["text"]
        
        # map token ids to word (whitespace) token ids
        predstrings = []
        for span in all_span:
            span_start  = span[0]
            span_end    = span[1]
            before      = text[:span_start]
            token_start = len(before.split())
            if len(before) == 0:    token_start = 0
            elif before[-1] != ' ': token_start -= 1

            num_tkns   = len(text[span_start:span_end+1].split())
            tkns       = [str(x) for x in range(token_start, token_start+num_tkns)]
            predstring = ' '.join(tkns)
            predstrings.append(predstring)

        #### Geclassate Record format 
        row = {  "text": text, "class_list": []}
        es = []
        for class_type, span, predstring in zip(classes, all_span, predstrings):
            if class_type!='Other':
              e = {
                'type' : class_type,
                'value': predstring,
                'start': span[0],
                'end'  : span[1],
                'text' : text[span[0]:span[1]]
              }
              es.append(e)
        row["class_list"] = es
    
        return row


    #### Convert
    pred_class = df_val[col_classlist].values
    valid      = df_val[[col_text]]
    valid['offset_mapping'] = offset_mapping
    valid = valid.to_dict(orient="records")

    ### pred_class : tuple(start, end, string)
    predicts= [pred2span(pred_class[i], valid[i]) for i in range(len(valid))]

    return [row['class_list'] for row in predicts]


    #pd_to_file( predicts, dirout + '/predict_class_visualize.parquet' )
    # pred = pd.read_csv("nguyen/20240215_171305/predict.csv")
    # pred
    # pred = pred[["text", "class_list"]]
    # eval(pred.class_tag.iloc[0])
    # pred["class_list"] = pred["class_list"].apply(eval)






###################################################################################################
if __name__ == "__main__":
    import fire
    fire.Fire()




""" logs 




(py39) (py39) \W$       pyclass run_infer_batch --dirmodel $dirmodel  --dirdata "ztmp/data/cats/arxiv/val/*"  --dirout ztmp/out/class/deberta_arxiv/

Config: Using /Users/macair/conda3/envs/py39/lib/python3.9/site-packages/utilmy/configs/myconfig/config.yaml
Config: Loading  /Users/macair/conda3/envs/py39/lib/python3.9/site-packages/utilmy/configs/myconfig/config.yaml
Config: Cannot read file /Users/macair/conda3/envs/py39/lib/python3.9/site-packages/utilmy/configs/myconfig/config.yaml 'str' object has no attribute 'suffix'
Config: Using default config
{'field1': 'test', 'field2': {'version': '1.0'}}

### Inference config   #####################################################
Config: Loading  ztmp/exp/20240525/174321-class_deberta-6/model/meta.json
{'model_name': 'microsoft/deberta-v3-base', 'dataloader_name': 'data_arxiv_load'

### Model Load  ############################################################
<transformers.pipelines.text_classification.TextClassificationPipeline object at

### Start Inference loop  ############################################################
nFiles:  1
0 ztmp/data/cats/arxiv/val/df_2000.parquet (10, 10)
ztmp/out/class/deberta_arxiv//df_pred_labels_0.parquet
                                           abstracts  ...                             pred_list_records
0  We present DeepMVI, a deep learning method for...  ...  {'label': 'DC', 'score': 0.5918803811073303}
1  We introduce Neural Contextual Anomaly Detecti...  ...  {'label': 'DC', 'score': 0.5875549912452698}
2  An innovations sequence of a time series is a ...  ...  {'label': 'DC', 'score': 0.5880179405212402}
3  In order to mitigate the spread of COVID-19, W...  ...  {'label': 'DC', 'score': 0.5901156663894653}
4  While previous distribution shift detection ap...  ...  {'label': 'DC', 'score': 0.5917407274246216}
5  We predict the emergence of extreme events in ...  ...  {'label': 'DC', 'score': 0.5919176340103149}
6  The spread of COVID-19 has coincided with the ...  ...  {'label': 'DC', 'score': 0.5906561017036438}
7  Until recently, the most accurate methods for ...  ...  {'label': 'DC', 'score': 0.5906267762184143}
8  Intelligent diagnosis method based on data-dri...  ...  {'label': 'DC', 'score': 0.5890864729881287}
9  Extracting the interaction rules of biological...  ...   {'label': 'DC', 'score': 0.590703547000885}

[10 rows x 11 columns]
(py39) (py39) \W$ 



"""







    
# Custom metric for evaluation
def zz_compute_metrics(eval_pred):
    ### not working
    logits = eval_pred.predictions
    labels = eval_pred.label_ids
    predictions = (torch.sigmoid(torch.Tensor(logits)) > 0.5).int()
    f1 = f1_score(torch.Tensor(labels).int().cpu().numpy(), predictions.cpu().numpy(), average='samples')
    return {"f1": f1}

def zz_metrics_calc_callback_train_v2(eval_pred):
    """  Durinh HugginFace training 
        Not working with evaluate

            import evaluate
            import numpy as np

            clf_metrics = evaluate.combine(["accuracy", "f1", "precision", "recall"])

            def sigmoid(x):
               return 1/(1 + np.exp(-x))

        def compute_metrics(eval_pred):

            predictions, labels = eval_pred
            predictions = sigmoid(predictions)
            predictions = (predictions > 0.5).astype(int).reshape(-1)
            return clf_metrics.compute(predictions=predictions, references=labels.astype(int).reshape(-1))
            references=labels.astype(int).reshape(-1))

    """
    # global L2I, I2L

    # metric = datasets.load_metric("seqeval")
    metric = evaluate.combine(["accuracy", "f1", "precision", "recall"])
    # pred_proba, labels = xtuple

    pred_logits = eval_pred.predictions
    labels     = eval_pred.label_ids

    # preds = np.argmax(preds_proba, axis=-1)
    pred_labels = (torch.sigmoid(torch.Tensor(pred_logits)) > 0.5).int()

    pred_labels =  pred_labels.detach().cpu().numpy()
    labels      = np.array(labels, dtype="int32") # [ [int(x) for x in vlist ] for vlist  in labels ]

    #####
    pred_labels = pred_labels.tolist()
    labels      = labels.tolist()

    # Remove ignored index (special tokens)
    #true_preds  =  [I2L[p ] for (p, l) in zip(preds, labels) if l != -100]
    #true_labels =  [I2L[l ] for (_, l) in zip(preds, labels) if l != -100] 
    # results = metric.compute(predictions=true_preds, references=true_labels)

    results = metric.compute(predictions= pred_labels, references= labels,
                             average='weighted')

    return {
        "precision": results["overall_precision"],
        "recall"   : results["overall_recall"],
        "f1"       : results["overall_f1"],
        "accuracy" : results["overall_accuracy"],
    }






def zz_data_tokenize_split_v2(data:dict, max_length=128, model_id="microsoft/deberta-base",):
    # Tokenization
    tokenizer = DebertaTokenizer.from_pretrained(model_id)

    def preprocess_function(examples):
        output = tokenizer(examples["text"], truncation=True, padding=True, max_length= max_length)
        # output["input_ids"] = output.pop("input_ids")  # Add input_ids to  output
        return output

    # Load  dataset
    ds = Dataset.from_dict(data)

    # Encode texts
    ds = ds.map(preprocess_function, batched=True)

    # Remove labels with only a single instance
    label_counts = Counter([tuple(label) for label in ds["labels"]])
    valid_labels = [label for label, count in label_counts.items() if count > 1]
    ds           = ds.filter(lambda example: tuple(example["labels"]) in valid_labels)

    # Split dataset into training and validation sets with stratification
    text_train, text_test, labels_train, labels_test = train_test_split(
        ds["text"],
        ds["labels"],
        test_size=0.2,
        random_state=42,
        stratify=ds["labels"],
    )

    train_dataset = Dataset.from_dict(
        {   "text": text_train,
            "labels": labels_train,
            "input_ids": ds["input_ids"][: len(text_train)],
            "attention_mask": ds["attention_mask"][: len(text_train)],
        }
    )

    test_dataset = Dataset.from_dict(
        {   "text": text_test,
            "labels": labels_test,
            "input_ids": ds["input_ids"][len(text_train) :],
            "attention_mask": ds["attention_mask"][len(text_train) :],
        }
    )
    return train_dataset, test_dataset, tokenizer

